import sentry_sdk
from flask import Flask, jsonify, request
from flask_cors import CORS
from flask_limiter import Limiter
import werkzeug
from webargs import fields, validate
from webargs.flaskparser import use_args
from util import util
from api import accounts, users, idps, root

sentry_sdk.init(
    dsn="https://e5b27a6e04405e77b14f51631b2c96f3@o4508066290532352.ingest.de.sentry.io/4508066432876624",
    traces_sample_rate=1.0,
    profiles_sample_rate=1.0,
)

app = Flask(__name__)
CORS(app)
limiter = Limiter(util.limit_by_user, app=app, default_limits=["20 per minute"])


@app.errorhandler(werkzeug.exceptions.TooManyRequests)
def handle_429(e):
    return jsonify(
        {
            "message": "You're making too many requests. Please wait for a few minutes before trying again.",
            "Allowed limit": e.description,
        }
    ), 429


@app.errorhandler(werkzeug.exceptions.BadRequest)
def handle_bad_request(e):
    return jsonify({"message": e.description}), 400


@app.errorhandler(werkzeug.exceptions.Unauthorized)
def handle_not_authorized(e):
    return jsonify({"message": e.description}), 401


@app.errorhandler(werkzeug.exceptions.Forbidden)
def handle_forbidden(e):
    return jsonify({"message": e.description}), 403


@app.errorhandler(werkzeug.exceptions.NotFound)
def handle_not_found(e):
    return jsonify({"message": e.description}), 404


@app.errorhandler(werkzeug.exceptions.UnprocessableEntity)
def handle_unprocessable_entity(e):
    validation_errors = e.data.get("messages")
    message = ""

    def build_message(message, d):
        if not d:
            return message
        for key in d:
            if isinstance(d[key], dict):
                message += f"""{str(key)}: """
                return build_message(message, d[key])
            elif isinstance(d[key], list):
                message += f"""{str(key)}: {", ".join(d[key])}\n"""
        return message

    if validation_errors:
        message = build_message("", validation_errors.get("json"))
    return jsonify(
        {
            "message": message,
            "validations": validation_errors,
        }
    ), 422


# ACCOUNTS


@app.route("/accounts", methods=["POST"])
@limiter.limit("5 per minute", key_func=util.limit_by_client, methods=["POST"])
@use_args(
    {
        "email": fields.Email(required=True),
        "firstName": fields.Str(required=True, validate=validate.Length(min=2)),
        "lastName": fields.Str(required=True, validate=validate.Length(min=2)),
        "password": fields.Str(required=True, validate=validate.Length(min=8)),
        "idpsToClaim": fields.List(fields.Str(), missing=[], allow_none=True),
    }
)
def register(args):
    return util.jsonify(accounts.create(args))


@app.route("/accounts/enrol", methods=["POST"])
@limiter.limit("5 per minute", key_func=util.limit_by_client, methods=["POST"])
@use_args({"token": fields.Str(), "password": fields.Str(required=True)})
def enrol(args):
    return util.jsonify(accounts.enrol(args))


@app.route("/accounts/sessions", methods=["POST"])
@limiter.limit("5 per minute", key_func=util.limit_by_client, methods=["POST"])
@use_args(
    {
        "email": fields.Email(required=True),
        "password": fields.Str(required=True),
        "idpsToClaim": fields.List(fields.Str(), missing=[], allow_none=True),
    }
)
def login(args):
    return util.jsonify(accounts.login(args))


@app.route("/accounts/sessions", methods=["DELETE"])
def logout():
    return util.jsonify(accounts.logout(util.get_user()))


@app.route("/accounts", methods=["DELETE"])
@use_args(
    {
        "password": fields.Str(required=True),
    }
)
def delete_account(args):
    return util.jsonify(accounts.delete(util.get_user(), args))


@app.route("/accounts/password", methods=["PUT"])
@limiter.limit("5 per minute", key_func=util.limit_by_user, methods=["PUT"])
@use_args(
    {
        "newPassword": fields.Str(required=True, validate=validate.Length(min=8)),
        "currentPassword": fields.Str(alow_none=True),
        "token": fields.Str(allow_none=True),
    }
)
def password(args):
    return util.jsonify(accounts.update_password(util.get_user(required=False), args))


@app.route("/accounts/password/reset", methods=["POST"])
@limiter.limit("5 per minute", key_func=util.limit_by_client, methods=["POST"])
@use_args(
    {
        "email": fields.Email(required=True),
    }
)
def reset_password(args):
    return util.jsonify(accounts.reset_password(args))


# Users


@app.route("/users/me", methods=["GET"])
def users_me():
    return util.jsonify(users.me(util.get_user()))


@app.route("/users/<id>", methods=["PUT"])
@use_args(
    {
        "firstName": fields.Str(validate=validate.Length(min=2)),
        "lastName": fields.Str(validate=validate.Length(min=2)),
        "email": fields.Email(),
    }
)
def user_route(args, id):
    return util.jsonify(users.update(util.get_user(), id, args))


# IDPs


@app.route("/idps", methods=["GET"])
@use_args({"include": fields.Str(missing="")}, location="query")
def idps_route_get(args):
    return util.jsonify(idps.get(util.get_user(required=False), args.get("include")))


@app.route("/idps", methods=["POST"])
@use_args(
    {
        "code": fields.Str(required=True, validate=validate.Length(min=2)),
        "name": fields.Str(required=True, validate=validate.Length(min=2)),
    }
)
def idps_route_post(args):
    return util.jsonify(idps.create(util.get_user(required=False), args))


@app.route("/idps/<id>", methods=["GET", "DELETE"])
def idp_route(id):
    if request.method == "GET":
        return util.jsonify(idps.get_one(util.get_user(required=False), id))
    if request.method == "DELETE":
        return util.jsonify(idps.delete(util.get_user(required=False), id))


@app.route("/idps/<id>", methods=["PUT"])
@use_args(
    {
        "code": fields.Str(validate=validate.Length(min=2)),
        "name": fields.Str(validate=validate.Length(min=2)),
    }
)
def idp_route_put(args, id):
    return util.jsonify(idps.update(util.get_user(required=False), id, args))


@app.route("/idps/<id>/sps", methods=["GET"])
def idp_sps_route_get(id):
    return util.jsonify(idps.get_sps(util.get_user(required=False), id))


@app.route("/idps/<id>/sps", methods=["POST"])
@use_args(
    {
        "name": fields.Str(required=True, validate=validate.Length(min=2)),
        "entityId": fields.Str(),
        "serviceUrl": fields.Str(),
        "callbackUrl": fields.Str(),
        "logoutUrl": fields.Str(),
        "logoutCallbackUrl": fields.Str(),
        "oauth2RedirectUri": fields.Str(),
    }
)
def idp_sps_route_post(args, id):
    return util.jsonify(idps.create_sp(util.get_user(required=False), id, args))


@app.route("/idps/<id>/sps/<sp_id>", methods=["PUT"])
@use_args(
    {
        "name": fields.Str(required=True, validate=validate.Length(min=2)),
        "entityId": fields.Str(),
        "serviceUrl": fields.Str(),
        "callbackUrl": fields.Str(),
        "logoutUrl": fields.Str(),
        "logoutCallbackUrl": fields.Str(),
        "oauth2RedirectUri": fields.Str(),
    }
)
def idp_sp_route_put(args, id, sp_id):
    return util.jsonify(idps.update_sp(util.get_user(required=False), id, sp_id, args))


@app.route("/idps/<id>/sps/<sp_id>", methods=["DELETE"])
def idp_sp_route(id, sp_id):
    return util.jsonify(idps.delete_sp(util.get_user(required=False), id, sp_id))


@app.route("/idps/<id>/users", methods=["GET"])
def idp_users_route_get(id):
    return util.jsonify(idps.get_users(util.get_user(required=False), id))


@app.route("/idps/<id>/users", methods=["POST"])
@use_args(
    {
        "email": fields.Email(required=True),
        "firstName": fields.Str(),
        "lastName": fields.Str(),
        "password": fields.Str(required=True, validate=validate.Length(min=3)),
        "attributes": fields.Dict(),
    }
)
def idp_users_route_post(args, id):
    return util.jsonify(idps.create_user(util.get_user(required=False), id, args))


@app.route("/idps/<id>/users/<user_id>", methods=["PUT"])
@use_args(
    {
        "email": fields.Email(required=True),
        "firstName": fields.Str(),
        "lastName": fields.Str(),
        "password": fields.Str(),
        "attributes": fields.Dict(),
    }
)
def idp_user_route_put(args, id, user_id):
    return util.jsonify(
        idps.update_user(util.get_user(required=False), id, user_id, args)
    )


@app.route("/idps/<id>/users/<user_id>", methods=["DELETE"])
def idp_user_route_delete(id, user_id):
    return util.jsonify(idps.delete_user(util.get_user(required=False), id, user_id))


@app.route("/idps/<id>/attributes", methods=["GET"])
def idp_attributes_route_get(id):
    return util.jsonify(idps.get_attributes(util.get_user(required=False), id))


@app.route("/idps/<id>/attributes", methods=["POST"])
@use_args(
    {
        "name": fields.Str(required=True, validate=validate.Length(min=2)),
        "defaultValue": fields.Str(),
        "samlMapping": fields.Str(),
    }
)
def idp_attributes_route_post(args, id):
    return util.jsonify(idps.create_attribute(util.get_user(required=False), id, args))


@app.route("/idps/<id>/attributes/<attr_id>", methods=["PUT"])
@use_args(
    {
        "name": fields.Str(required=True, validate=validate.Length(min=2)),
        "defaultValue": fields.Str(),
        "samlMapping": fields.Str(),
    }
)
def idp_attribute_route_put(args, id, attr_id):
    return util.jsonify(
        idps.update_attribute(util.get_user(required=False), id, attr_id, args)
    )


@app.route("/idps/<id>/attributes/<attr_id>", methods=["DELETE"])
def idp_attribute_route_delete(id, attr_id):
    return util.jsonify(
        idps.delete_attribute(util.get_user(required=False), id, attr_id)
    )


@app.route("/idps/<id>/saml2/logs", methods=["GET"])
def idp_saml_logs_route(id):
    return util.jsonify(idps.get_logs(util.get_user(required=False), id))


@app.route("/idps/<id>/oauth2/logs", methods=["GET"])
def idp_oauth_logs_route(id):
    return util.jsonify(idps.get_oauth_logs(util.get_user(required=False), id))


# Root


@app.route("/root/users", methods=["GET"])
def root_users():
    return util.jsonify(root.get_users(util.get_user()))
